############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
from __future__ import annotations

import asyncio
import logging
import sys
import typing
from threading import Event

from pygls.exceptions import JsonRpcException, PyglsError
from pygls.io_ import run_async, run_websocket
from pygls.protocol import JsonRPCProtocol, default_converter

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import List
    from typing import Optional
    from typing import Type

    from cattrs import Converter

logger = logging.getLogger(__name__)


class JsonRPCClient:
    """Base JSON-RPC client."""

    def __init__(
        self,
        protocol_cls: Type[JsonRPCProtocol] = JsonRPCProtocol,
        converter_factory: Callable[[], Converter] = default_converter,
    ):
        # Strictly speaking, `JsonRPCProtocol` wants a `JsonRPCServer`, not a
        # `JsonRPCClient`. However they're similar enough for our purposes, which
        # is that this client will mostly be used in testing contexts.
        self.protocol = protocol_cls(self, converter_factory())  # type: ignore

        self._server: Optional[asyncio.subprocess.Process] = None
        self._stop_event = Event()
        self._async_tasks: List[asyncio.Task[Any]] = []

    @property
    def stopped(self) -> bool:
        """Return ``True`` if the client has been stopped."""
        return self._stop_event.is_set()

    def feature(
        self,
        feature_name: str,
        options: Optional[Any] = None,
    ):
        """Decorator used to register LSP features.

        Example
        -------
        ::

           import logging
           from pygls.client import JsonRPCClient

           ls = JsonRPCClient()

           @ls.feature('window/logMessage')
           def completions(ls, params):
               logging.info("%s", params.message)
        """
        return self.protocol.fm.feature(feature_name, options)

    async def start_io(self, cmd: str, *args, **kwargs):
        """Start the given server and communicate with it over stdio."""

        logger.debug("Starting server process: %s", " ".join([cmd, *args]))
        server = await asyncio.create_subprocess_exec(
            cmd,
            *args,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            **kwargs,
        )

        # Keep mypy happy
        if server.stdout is None:
            raise RuntimeError("Server process is missing a stdout stream")

        # Keep mypy happy
        if server.stdin is None:
            raise RuntimeError("Server process is missing a stdin stream")

        self.protocol.set_writer(server.stdin)
        connection = asyncio.create_task(
            run_async(
                stop_event=self._stop_event,
                reader=server.stdout,
                protocol=self.protocol,
                logger=logger,
                error_handler=self.report_server_error,
            )
        )
        notify_exit = asyncio.create_task(self._server_exit())

        self._server = server
        self._async_tasks.extend([connection, notify_exit])

    async def start_tcp(self, host: str, port: int):
        """Start communicating with a server over TCP."""
        reader, writer = await asyncio.open_connection(host, port)

        self.protocol.set_writer(writer)
        connection = asyncio.create_task(
            run_async(
                stop_event=self._stop_event,
                reader=reader,
                protocol=self.protocol,
                logger=logger,
                error_handler=self.report_server_error,
            )
        )

        self._async_tasks.extend([connection])

    async def start_ws(self, host: str, port: int):
        """Start communicating with a server over WebSockets."""

        try:
            from websockets.asyncio.client import connect
        except ImportError:
            logger.exception(
                "Run `pip install pygls[ws]` to install dependencies required for websockets."
            )
            sys.exit(1)

        uri = f"ws://{host}:{port}"
        websocket = await connect(uri)
        connection = asyncio.create_task(
            run_websocket(
                stop_event=self._stop_event,
                websocket=websocket,
                protocol=self.protocol,
                logger=logger,
                error_handler=self.report_server_error,
            )
        )
        self._async_tasks.extend([connection])

        # Yield control to the event loop, gives the run_websocket task chance to spin up.
        await asyncio.sleep(0)

    async def _server_exit(self):
        """Cleanup handler that runs when the server process managed by the client exits"""
        if self._server is None:
            return

        await self._server.wait()

        pid = self._server.pid
        returncode = self._server.returncode

        reason = f"Server process {pid} exited with return code: {returncode}"
        logger.debug(reason)

        # Cancel any pending requests
        for id_, fut in self.protocol._request_futures.items():
            if not fut.done():
                fut.set_exception(RuntimeError(reason))
                logger.debug("Cancelled pending request '%s': %s", id_, reason)

        try:
            await self.server_exit(self._server)
        except Exception:
            logger.exception("Error in server_exit handler")

        self._stop_event.set()

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the server process exits."""

    def _report_server_error(
        self, error: Exception, source: type[PyglsError] | type[JsonRpcException]
    ):
        try:
            self.report_server_error(error, source)
        except Exception:
            logger.exception("Unable to report error")

    def report_server_error(
        self, error: Exception, source: type[PyglsError] | type[JsonRpcException]
    ):
        """Called when the server does something unexpected e.g. respond with malformed
        JSON."""

    async def stop(self):
        self._stop_event.set()

        if self._server is not None and self._server.returncode is None:
            await self._server.wait()

        if len(self._async_tasks) > 0:
            await asyncio.gather(*self._async_tasks)



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################

# GENERATED FROM scripts/generate_code.py -- DO NOT EDIT
# flake8: noqa
from __future__ import annotations

from lsprotocol import types
from pygls.protocol import LanguageServerProtocol
from pygls.protocol import default_converter
from pygls.server import JsonRPCServer
import typing

if typing.TYPE_CHECKING:
    from cattrs import Converter
    from concurrent.futures import Future
    from typing import Any
    from typing import Callable
    from typing import Optional
    from typing import Sequence


class BaseLanguageServer(JsonRPCServer):

    protocol: LanguageServerProtocol

    def __init__(
        self,
        protocol_cls: type[LanguageServerProtocol] = LanguageServerProtocol,
        converter_factory: Callable[[], Converter] = default_converter,
        max_workers: int | None = None,
    ):
        super().__init__(protocol_cls, converter_factory, max_workers)

    def client_register_capability(
        self,
        params: types.RegistrationParams,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`client/registerCapability` request.

        The `client/registerCapability` request is sent from the server to the client to register a new capability
        handler on the client side.
        """
        return self.protocol.send_request("client/registerCapability", params, callback)

    async def client_register_capability_async(
        self,
        params: types.RegistrationParams,
    ) -> None:
        """Make a :lsp:`client/registerCapability` request.

        The `client/registerCapability` request is sent from the server to the client to register a new capability
        handler on the client side.
        """
        return await self.protocol.send_request_async("client/registerCapability", params)

    def client_unregister_capability(
        self,
        params: types.UnregistrationParams,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`client/unregisterCapability` request.

        The `client/unregisterCapability` request is sent from the server to the client to unregister a previously registered capability
        handler on the client side.
        """
        return self.protocol.send_request("client/unregisterCapability", params, callback)

    async def client_unregister_capability_async(
        self,
        params: types.UnregistrationParams,
    ) -> None:
        """Make a :lsp:`client/unregisterCapability` request.

        The `client/unregisterCapability` request is sent from the server to the client to unregister a previously registered capability
        handler on the client side.
        """
        return await self.protocol.send_request_async("client/unregisterCapability", params)

    def window_show_document(
        self,
        params: types.ShowDocumentParams,
        callback: Optional[Callable[[types.ShowDocumentResult], None]] = None,
    ) -> Future[types.ShowDocumentResult]:
        """Make a :lsp:`window/showDocument` request.

        A request to show a document. This request might open an
        external program depending on the value of the URI to open.
        For example a request to open `https://code.visualstudio.com/`
        will very likely open the URI in a WEB browser.

        @since 3.16.0
        """
        return self.protocol.send_request("window/showDocument", params, callback)

    async def window_show_document_async(
        self,
        params: types.ShowDocumentParams,
    ) -> types.ShowDocumentResult:
        """Make a :lsp:`window/showDocument` request.

        A request to show a document. This request might open an
        external program depending on the value of the URI to open.
        For example a request to open `https://code.visualstudio.com/`
        will very likely open the URI in a WEB browser.

        @since 3.16.0
        """
        return await self.protocol.send_request_async("window/showDocument", params)

    def window_show_message_request(
        self,
        params: types.ShowMessageRequestParams,
        callback: Optional[Callable[[Optional[types.MessageActionItem]], None]] = None,
    ) -> Future[Optional[types.MessageActionItem]]:
        """Make a :lsp:`window/showMessageRequest` request.

        The show message request is sent from the server to the client to show a message
        and a set of options actions to the user.
        """
        return self.protocol.send_request("window/showMessageRequest", params, callback)

    async def window_show_message_request_async(
        self,
        params: types.ShowMessageRequestParams,
    ) -> Optional[types.MessageActionItem]:
        """Make a :lsp:`window/showMessageRequest` request.

        The show message request is sent from the server to the client to show a message
        and a set of options actions to the user.
        """
        return await self.protocol.send_request_async("window/showMessageRequest", params)

    def window_work_done_progress_create(
        self,
        params: types.WorkDoneProgressCreateParams,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`window/workDoneProgress/create` request.

        The `window/workDoneProgress/create` request is sent from the server to the client to initiate progress
        reporting from the server.
        """
        return self.protocol.send_request("window/workDoneProgress/create", params, callback)

    async def window_work_done_progress_create_async(
        self,
        params: types.WorkDoneProgressCreateParams,
    ) -> None:
        """Make a :lsp:`window/workDoneProgress/create` request.

        The `window/workDoneProgress/create` request is sent from the server to the client to initiate progress
        reporting from the server.
        """
        return await self.protocol.send_request_async("window/workDoneProgress/create", params)

    def workspace_apply_edit(
        self,
        params: types.ApplyWorkspaceEditParams,
        callback: Optional[Callable[[types.ApplyWorkspaceEditResult], None]] = None,
    ) -> Future[types.ApplyWorkspaceEditResult]:
        """Make a :lsp:`workspace/applyEdit` request.

        A request sent from the server to the client to modified certain resources.
        """
        return self.protocol.send_request("workspace/applyEdit", params, callback)

    async def workspace_apply_edit_async(
        self,
        params: types.ApplyWorkspaceEditParams,
    ) -> types.ApplyWorkspaceEditResult:
        """Make a :lsp:`workspace/applyEdit` request.

        A request sent from the server to the client to modified certain resources.
        """
        return await self.protocol.send_request_async("workspace/applyEdit", params)

    def workspace_code_lens_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/codeLens/refresh` request.

        A request to refresh all code actions

        @since 3.16.0
        """
        return self.protocol.send_request("workspace/codeLens/refresh", params, callback)

    async def workspace_code_lens_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/codeLens/refresh` request.

        A request to refresh all code actions

        @since 3.16.0
        """
        return await self.protocol.send_request_async("workspace/codeLens/refresh", params)

    def workspace_configuration(
        self,
        params: types.ConfigurationParams,
        callback: Optional[Callable[[Sequence[Optional[Any]]], None]] = None,
    ) -> Future[Sequence[Optional[Any]]]:
        """Make a :lsp:`workspace/configuration` request.

        The 'workspace/configuration' request is sent from the server to the client to fetch a certain
        configuration setting.

        This pull model replaces the old push model were the client signaled configuration change via an
        event. If the server still needs to react to configuration changes (since the server caches the
        result of `workspace/configuration` requests) the server should register for an empty configuration
        change event and empty the cache if such an event is received.
        """
        return self.protocol.send_request("workspace/configuration", params, callback)

    async def workspace_configuration_async(
        self,
        params: types.ConfigurationParams,
    ) -> Sequence[Optional[Any]]:
        """Make a :lsp:`workspace/configuration` request.

        The 'workspace/configuration' request is sent from the server to the client to fetch a certain
        configuration setting.

        This pull model replaces the old push model were the client signaled configuration change via an
        event. If the server still needs to react to configuration changes (since the server caches the
        result of `workspace/configuration` requests) the server should register for an empty configuration
        change event and empty the cache if such an event is received.
        """
        return await self.protocol.send_request_async("workspace/configuration", params)

    def workspace_diagnostic_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/diagnostic/refresh` request.

        The diagnostic refresh request definition.

        @since 3.17.0
        """
        return self.protocol.send_request("workspace/diagnostic/refresh", params, callback)

    async def workspace_diagnostic_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/diagnostic/refresh` request.

        The diagnostic refresh request definition.

        @since 3.17.0
        """
        return await self.protocol.send_request_async("workspace/diagnostic/refresh", params)

    def workspace_folding_range_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/foldingRange/refresh` request.

        @since 3.18.0
        @proposed
        """
        return self.protocol.send_request("workspace/foldingRange/refresh", params, callback)

    async def workspace_folding_range_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/foldingRange/refresh` request.

        @since 3.18.0
        @proposed
        """
        return await self.protocol.send_request_async("workspace/foldingRange/refresh", params)

    def workspace_inlay_hint_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/inlayHint/refresh` request.

        @since 3.17.0
        """
        return self.protocol.send_request("workspace/inlayHint/refresh", params, callback)

    async def workspace_inlay_hint_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/inlayHint/refresh` request.

        @since 3.17.0
        """
        return await self.protocol.send_request_async("workspace/inlayHint/refresh", params)

    def workspace_inline_value_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/inlineValue/refresh` request.

        @since 3.17.0
        """
        return self.protocol.send_request("workspace/inlineValue/refresh", params, callback)

    async def workspace_inline_value_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/inlineValue/refresh` request.

        @since 3.17.0
        """
        return await self.protocol.send_request_async("workspace/inlineValue/refresh", params)

    def workspace_semantic_tokens_refresh(
        self,
        params: None,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/semanticTokens/refresh` request.

        @since 3.16.0
        """
        return self.protocol.send_request("workspace/semanticTokens/refresh", params, callback)

    async def workspace_semantic_tokens_refresh_async(
        self,
        params: None,
    ) -> None:
        """Make a :lsp:`workspace/semanticTokens/refresh` request.

        @since 3.16.0
        """
        return await self.protocol.send_request_async("workspace/semanticTokens/refresh", params)

    def workspace_text_document_content_refresh(
        self,
        params: types.TextDocumentContentRefreshParams,
        callback: Optional[Callable[[None], None]] = None,
    ) -> Future[None]:
        """Make a :lsp:`workspace/textDocumentContent/refresh` request.

        The `workspace/textDocumentContent` request is sent from the server to the client to refresh
        the content of a specific text document.

        @since 3.18.0
        @proposed
        """
        return self.protocol.send_request("workspace/textDocumentContent/refresh", params, callback)

    async def workspace_text_document_content_refresh_async(
        self,
        params: types.TextDocumentContentRefreshParams,
    ) -> None:
        """Make a :lsp:`workspace/textDocumentContent/refresh` request.

        The `workspace/textDocumentContent` request is sent from the server to the client to refresh
        the content of a specific text document.

        @since 3.18.0
        @proposed
        """
        return await self.protocol.send_request_async("workspace/textDocumentContent/refresh", params)

    def workspace_workspace_folders(
        self,
        params: None,
        callback: Optional[Callable[[Optional[Sequence[types.WorkspaceFolder]]], None]] = None,
    ) -> Future[Optional[Sequence[types.WorkspaceFolder]]]:
        """Make a :lsp:`workspace/workspaceFolders` request.

        The `workspace/workspaceFolders` is sent from the server to the client to fetch the open workspace folders.
        """
        return self.protocol.send_request("workspace/workspaceFolders", params, callback)

    async def workspace_workspace_folders_async(
        self,
        params: None,
    ) -> Optional[Sequence[types.WorkspaceFolder]]:
        """Make a :lsp:`workspace/workspaceFolders` request.

        The `workspace/workspaceFolders` is sent from the server to the client to fetch the open workspace folders.
        """
        return await self.protocol.send_request_async("workspace/workspaceFolders", params)

    def cancel_request(self, params: types.CancelParams) -> None:
        """Send a :lsp:`$/cancelRequest` notification.


        """
        self.protocol.notify("$/cancelRequest", params)

    def log_trace(self, params: types.LogTraceParams) -> None:
        """Send a :lsp:`$/logTrace` notification.


        """
        self.protocol.notify("$/logTrace", params)

    def progress(self, params: types.ProgressParams) -> None:
        """Send a :lsp:`$/progress` notification.


        """
        self.protocol.notify("$/progress", params)

    def telemetry_event(self, params: typing.Optional[typing.Any]) -> None:
        """Send a :lsp:`telemetry/event` notification.

        The telemetry event notification is sent from the server to the client to ask
        the client to log telemetry data.
        """
        self.protocol.notify("telemetry/event", params)

    def text_document_publish_diagnostics(self, params: types.PublishDiagnosticsParams) -> None:
        """Send a :lsp:`textDocument/publishDiagnostics` notification.

        Diagnostics notification are sent from the server to the client to signal
        results of validation runs.
        """
        self.protocol.notify("textDocument/publishDiagnostics", params)

    def window_log_message(self, params: types.LogMessageParams) -> None:
        """Send a :lsp:`window/logMessage` notification.

        The log message notification is sent from the server to the client to ask
        the client to log a particular message.
        """
        self.protocol.notify("window/logMessage", params)

    def window_show_message(self, params: types.ShowMessageParams) -> None:
        """Send a :lsp:`window/showMessage` notification.

        The show message notification is sent from a server to a client to ask
        the client to display a particular message in the user interface.
        """
        self.protocol.notify("window/showMessage", params)



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
from __future__ import annotations

import asyncio
import logging
import sys
import typing
from concurrent.futures import ThreadPoolExecutor
from threading import Event

import cattrs

from pygls import IS_WASM
from pygls.exceptions import JsonRpcException, PyglsError
from pygls.io_ import StdinAsyncReader, StdoutWriter, run, run_async, run_websocket
from pygls.protocol import JsonRPCProtocol

if typing.TYPE_CHECKING:
    from typing import Any, BinaryIO, Callable, Optional, Type, TypeVar, Union

    from websockets.asyncio.server import Server as WSServer
    from websockets.asyncio.server import ServerConnection

    F = TypeVar("F", bound=Callable)
    ServerErrors = Union[type[PyglsError], type[JsonRpcException]]


logger = logging.getLogger(__name__)


class JsonRPCServer:
    """Base server class

    Parameters
    ----------
    protocol_cls
       Protocol implementation that should derive from
       :class:`~pygls.protocol.JsonRPCProtocol`

    converter_factory
       Factory function to use when constructing a cattrs converter.

    max_workers
       Maximum number of workers for `ThreadPoolExecutor`

    """

    protocol: JsonRPCProtocol

    def __init__(
        self,
        protocol_cls: Type[JsonRPCProtocol],
        converter_factory: Callable[[], cattrs.Converter],
        max_workers: int | None = None,
    ):
        self._max_workers = max_workers
        self._server: asyncio.Server | WSServer | None = None
        self._stop_event: Event | None = None
        self._thread_pool: ThreadPoolExecutor | None = None

        self.protocol = protocol_cls(self, converter_factory())

    def shutdown(self):
        """Shutdown server."""
        logger.info("Shutting down the server")

        if self._stop_event is not None:
            self._stop_event.set()

        if self._thread_pool:
            self._thread_pool.shutdown()

        if self._server:
            self._server.close()

    def _report_server_error(
        self,
        error: Exception,
        source: ServerErrors,
    ):
        # Prevent recursive error reporting
        try:
            self.report_server_error(error, source)
        except Exception:
            logger.warning("Failed to report error")

    def report_server_error(self, error: Exception, source: ServerErrors):
        """Default error reporter."""
        logger.error("%s", error)

    def start_io(
        self, stdin: Optional[BinaryIO] = None, stdout: Optional[BinaryIO] = None
    ):
        """Starts an IO server."""

        if IS_WASM:
            self._start_io_sync(stdin, stdout)
        else:
            self._start_io_async(stdin, stdout)

    def _start_io_async(
        self, stdin: Optional[BinaryIO] = None, stdout: Optional[BinaryIO] = None
    ):
        """Starts an asynchronous IO server."""
        logger.info("Starting async IO server")

        self._stop_event = Event()
        reader = StdinAsyncReader(stdin or sys.stdin.buffer, self.thread_pool)
        writer = StdoutWriter(stdout or sys.stdout.buffer)
        self.protocol.set_writer(writer)

        try:
            asyncio.run(
                run_async(
                    stop_event=self._stop_event,
                    reader=reader,
                    protocol=self.protocol,
                    logger=logger,
                    error_handler=self.report_server_error,
                )
            )
        except BrokenPipeError:
            logger.error("Connection to the client is lost! Shutting down the server.")
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self.shutdown()

    def _start_io_sync(
        self, stdin: Optional[BinaryIO] = None, stdout: Optional[BinaryIO] = None
    ):
        """Starts an synchronous IO server."""
        logger.info("Starting sync IO server")

        self._stop_event = Event()
        writer = StdoutWriter(stdout or sys.stdout.buffer)
        self.protocol.set_writer(writer)

        try:
            asyncio.run(
                run(
                    stop_event=self._stop_event,
                    reader=stdin or sys.stdin.buffer,
                    protocol=self.protocol,
                    logger=logger,
                    error_handler=self.report_server_error,
                )
            )
        except BrokenPipeError:
            logger.error("Connection to the client is lost! Shutting down the server.")
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self.shutdown()

    def start_tcp(self, host: str, port: int) -> None:
        """Starts TCP server."""
        logger.info("Starting TCP server on %s:%s", host, port)

        self._stop_event = stop_event = Event()

        async def lsp_connection(
            reader: asyncio.StreamReader, writer: asyncio.StreamWriter
        ):
            logger.debug("Connected to client")
            self.protocol.set_writer(writer)  # type: ignore
            await run_async(
                stop_event=stop_event,
                reader=reader,
                protocol=self.protocol,
                logger=logger,
                error_handler=self.report_server_error,
            )
            logger.debug("Main loop finished")
            self.shutdown()

        async def tcp_server(h: str, p: int):
            self._server = await asyncio.start_server(lsp_connection, h, p)

            addrs = ", ".join(str(sock.getsockname()) for sock in self._server.sockets)
            logger.info(f"Serving on {addrs}")

            async with self._server:
                await self._server.serve_forever()

        try:
            asyncio.run(tcp_server(host, port))
        except asyncio.CancelledError:
            logger.debug("Server was cancelled")

    def start_ws(self, host: str, port: int) -> None:
        """Starts WebSocket server."""
        try:
            from websockets.asyncio.server import serve
        except ImportError:
            logger.error(
                "Run `pip install pygls[ws]` to install dependencies required for websockets."
            )
            sys.exit(1)

        logger.info("Starting WebSocket server on {}:{}".format(host, port))
        self._stop_event = stop_event = Event()

        async def lsp_connection(websocket: ServerConnection):
            await run_websocket(
                stop_event=stop_event,
                websocket=websocket,
                protocol=self.protocol,
                logger=logger,
                error_handler=self.report_server_error,
            )
            self.shutdown()

        async def ws_server(h: str, p: int):
            self._server = await serve(lsp_connection, host, port)

            addrs = ", ".join(str(sock.getsockname()) for sock in self._server.sockets)
            logger.info(f"Serving on {addrs}")

            async with self._server:
                await self._server.serve_forever()

        try:
            asyncio.run(ws_server(host, port))
        except asyncio.CancelledError:
            logger.debug("Server was cancelled")

    def command(self, command_name: str) -> Callable[[F], F]:
        """Decorator used to register custom commands.

        Example
        -------
        ::

           @ls.command('myCustomCommand')
           def my_cmd(ls, a, b, c):
               pass
        """
        return self.protocol.fm.command(command_name)

    def thread(self) -> Callable[[F], F]:
        """Decorator that mark function to execute it in a thread."""
        return self.protocol.fm.thread()

    def feature(
        self,
        feature_name: str,
        options: Any | None = None,
    ) -> Callable[[F], F]:
        """Decorator used to register LSP features.

        Example
        -------
        ::

           @ls.feature('textDocument/completion', CompletionOptions(trigger_characters=['.']))
           def completions(ls, params: CompletionParams):
               return CompletionList(is_incomplete=False, items=[CompletionItem("Completion 1")])
        """
        return self.protocol.fm.feature(feature_name, options)

    @property
    def thread_pool(self) -> ThreadPoolExecutor:
        """Returns thread pool instance (lazy initialization)."""
        if not self._thread_pool:
            self._thread_pool = ThreadPoolExecutor(max_workers=self._max_workers)

        return self._thread_pool
